# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

from task_processing.interfaces.runner import Runner


class Promise(Runner):
    def __init__(self, executor):
        pass

    def run(self, task_config):
        pass
